from __future__ import annotations

from datetime import date, timedelta

import pytest

from urelativedelta import relativedelta


def test_duration_arithmetic():
    x = relativedelta(months=5 * 12 + 7, seconds=100)
    y = relativedelta(months=3 * 12 + 6, seconds=300)
    z = timedelta(days=100)

    assert x + y == relativedelta(months=9 * 12 + 1, seconds=400)
    assert x - y == relativedelta(months=2 * 12 + 1, seconds=-200)
    assert x + z == relativedelta(months=5 * 12 + 7, days=100, seconds=100)
    assert y + x == y + x, "Addition should be symmetric"
    assert x - y == -(y - x), "Subtraction should be anti-symmetric"
    assert y + z == z + y, "Addition should be symmetric"
    assert y - z == -(z - y), "Subtraction should be anti-symmetric"

    assert x // 2 == relativedelta(months=5 * 6 + 3, seconds=50)
    assert x * 2 == relativedelta(months=10 * 12 + 14, seconds=200)

    # Dividing a 1-month by e.g. 2 has no unambiguous meaning
    with pytest.raises(TypeError, match="unsupported operand"):
        x / 2


def test_date_arithmetic():
    base = date(2020, 2, 29)

    assert base + relativedelta(months=24) == date(2022, 2, 28)
    assert base + relativedelta(months=48) == date(2024, 2, 29)

    assert base - relativedelta(months=24) == date(2018, 2, 28)
    assert base - relativedelta(months=48) == date(2016, 2, 29)

    not_leap = date(2020, 2, 28)
    tricky_delta = relativedelta(months=24, days=1)
    assert base + tricky_delta == date(2022, 3, 1)
    assert base + tricky_delta == not_leap + tricky_delta

    tricky_delta = relativedelta(months=24, days=-1)
    assert base - tricky_delta == date(2018, 3, 1)
    assert base - tricky_delta == not_leap - tricky_delta



from __future__ import annotations

import sys
from datetime import datetime, timedelta
from timeit import timeit

from dateutil.relativedelta import relativedelta as du_delta

from urelativedelta import relativedelta

klass = sys.argv[1]


dates = [datetime(2000, 1, 1) + timedelta(days=n) for n in range(5_000)]


def do_shifts():
    if klass == "relativedelta":
        for d in dates:
            d + relativedelta(years=100)
    if klass == "dateutil":
        for d in dates:
            d + du_delta(years=100)


print(klass, timeit(do_shifts, number=1000))



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import date, datetime
    from typing import TypeVar

    D = TypeVar("D", date, datetime)


def is_leap_year(year: int) -> bool:
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def normalise_day(year: int, month: int, day: int) -> int:
    """Shift the day backwards until it lies in the month.

    XXX: No attempt is made to handle days outside the 1-31 range.
    """
    if day <= 28:
        return day
    elif month == 2:
        return 28 + is_leap_year(year)
    elif day == 31 and (month == 4 or month == 6 or month == 9 or month == 11):
        return 30
    else:
        return day


def shift_months(date: D, months: int) -> D:
    """Shift a date by the given number of months.

    Ambiguous month-ends are shifted backwards as necessary."""
    year = date.year + (date.month + months - 1) // 12
    month = 1 + (date.month + months - 1) % 12
    day = normalise_day(year, month, date.day)

    return date.replace(year=year, month=month, day=day)


def shift_years(date: D, years: int) -> D:
    """Shift a date by the given number of years.

    Ambiguous month-ends are shifted backwards as necessary."""
    return shift_months(date, years * 12)


def with_day(date: D, day: int) -> D:
    """Shift the date to have the given day.

    Ambiguous month-ends are shifted backwards as necessary.
    """
    return date.replace(day=normalise_day(date.year, date.month, day))


def with_month(date: D, month: int) -> D:
    """Shift the date to have the given month.

    Ambiguous month-ends are shifted backwards as necessary.
    """
    if not 1 <= month <= 12:
        raise ValueError(f"month {month} should be between 1 and 12")

    return shift_months(date, month - date.month)


def with_year(date: D, year: int) -> D:
    """Shift the date to have the given year.

    Ambiguous month-ends are shifted backwards as necessary.
    """
    return shift_years(date, year - date.year)



"""Implements a relativedelta extending python's timedelta with months and years."""
from __future__ import annotations

from datetime import date as _date, datetime as _datetime, timedelta as _pytimedelta
from typing import TYPE_CHECKING

from .utils import shift_months

if TYPE_CHECKING:
    from typing import Any


class relativedelta:  # noqa: N801
    def __init__(self, years: int = 0, months: int = 0, timedelta=None, **delta_kwargs):
        self.months = months + 12 * years
        self.timedelta = _pytimedelta(**delta_kwargs)
        if timedelta is not None:
            self.timedelta += timedelta

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, relativedelta):
            return self.months == other.months and self.timedelta == other.timedelta
        elif isinstance(other, _pytimedelta):
            return self.months == 0 and self.timedelta == other
        else:
            return False

    def __hash__(self):
        return hash((self.months, self.timedelta))

    def __bool__(self) -> bool:
        return bool(self.months or self.timedelta)

    def __neg__(self) -> relativedelta:
        return type(self)(months=-self.months, timedelta=-self.timedelta)

    def __add__(self, other) -> relativedelta:
        if isinstance(other, relativedelta):
            months = self.months + other.months
            delta = self.timedelta + other.timedelta
        elif isinstance(other, _pytimedelta):
            months = self.months
            delta = self.timedelta + other
        else:
            return NotImplemented

        return self.__class__(months=months, timedelta=delta)

    def __sub__(self, other) -> relativedelta:
        return self + (-other)

    def __radd__(self, other):
        if isinstance(other, (_date, _datetime)):
            return shift_months(other, self.months) + self.timedelta
        elif isinstance(other, (_pytimedelta, relativedelta)):
            return self + other
        else:
            return NotImplemented

    def __rsub__(self, other):
        return other + (-self)

    def __mul__(self, n: int) -> relativedelta:
        return self.__class__(months=self.months * n, timedelta=self.timedelta * n)

    def __floordiv__(self, n: int) -> relativedelta:
        return self.__class__(months=self.months // n, timedelta=self.timedelta // n)

    def __repr__(self) -> str:
        return f"relativedelta(months={self.months}, timedelta={self.timedelta})"



from __future__ import annotations

from .relativedelta import relativedelta
from .utils import (
    is_leap_year,
    shift_months,
    shift_years,
    with_day,
    with_month,
    with_year,
)

__all__ = [
    "is_leap_year",
    "shift_months",
    "shift_years",
    "with_day",
    "with_month",
    "with_year",
    "relativedelta",
]
